"""
Module responsible for utilitarian decorators used to log the duration
of a given function execution.
"""
import asyncio
from contextlib import contextmanager
from functools import wraps
from time import perf_counter

from yajaw import YajawConfig


def duration(func):
    """
    duration Decorator used to log the duration of the target function.

    Decorated target functions with this one will capture the start and end
    times of its execution and print a log message informing the total
    duration in seconds. It uses perf_counter() from the time packge.
    It works for both sync and async functions.
    """

    @contextmanager
    def _wrapping_logic():
        "Main logic added by the decorator."
        start_ts = perf_counter()
        yield
        dur = perf_counter() - start_ts
        YajawConfig.LOGGER.info(f"{func.__name__} took {dur:.2f} seconds")

    @wraps(func)
    def _wrapper(*args, **kwargs):
        "Wrapper function that determines the right context for the wrapped function."
        if not asyncio.iscoroutinefunction(func):
            with _wrapping_logic():
                return func(*args, **kwargs)
        else:

            async def _async_wrapping():
                "Function used to wrap async functions to be decorated."
                with _wrapping_logic():
                    return await func(*args, **kwargs)

            return _async_wrapping()

    return _wrapper



# SPDX-FileCopyrightText: 2023-present rmrighes <rmrighes@users.noreply.github.com>
#
# SPDX-License-Identifier: MIT
"""
Module responsible for the initialization of configuration settings.
There's no need to import it directly.
"""

from enum import Enum

from yajaw.configuration import YajawConfig

__all__ = ["jira", "configuration", "exceptions", "ApiType"]


ApiType = Enum("API", ["CLASSIC", "AGILE", "INTERNAL"])
Option = Enum("Confirmation", ["YES", "NO"])

YajawConfig.load_initial_settings()



"""
Module responsible for handling the yajaw configuration.
"""
import asyncio
import logging
import tomllib
from pathlib import Path
from typing import ClassVar


class YajawConfig:
    """
     Class representing the configuration used by yajaw.

    It defines a series of class attributes representing the
    configuration settings used by yajaw. It also provides
    the basic load and update configuration settings methods.

    Attributes:
        JIRA_PAT: Personal access token used to authenticate in Jira
        JIRA_BASE_URL: Base url serving the Jira instance
        SERVER_API_V2: Path serving the rest server API version 2
        SERVER_API: Path serving the desired rest server API version
        AGILE_API_V1: Path serving the rest agile API version 1
        AGILE_API: Path serving the rest agile API version 1
        GREENHOPPER_API: Path serving the internal rest greenhopper API
        TRIES: How many time a request will be attempted before it fails
        DELAY: Number of seconds to wait before submitting the next request
        BACKOFF: Number multiplied against the delay to define its new value\
        in order to adjust the load against the Jira instance
        LOGGER: Logger instance created based on configuration settings
        SEMAPHORE: Semaphore object created based on configuration settings
        TIMEOUT: Number of seconds used to configure the semaphore timeout setting
        DEFAULT_PAGINATION: Initial dictionary with the start position and number \
        of results to be requested in paginated requests

    Raises:
        NameError: Raised when it can't update the configuration\
        using the provided section and settings.
    """

    JIRA_PAT: str
    JIRA_BASE_URL: str
    SERVER_API_V2: str
    SERVER_API: str
    AGILE_API_V1: str
    AGILE_API: str
    GREENHOPPER_API: str
    TRIES: int
    DELAY: float
    BACKOFF: float
    LOGGER: logging.Logger
    SEMAPHORE: asyncio.BoundedSemaphore
    TIMEOUT: int
    DEFAULT_PAGINATION: dict

    _MIN_SEMAPHORE_LIMIT: int = 5

    _configuration_settings: ClassVar = {
        "jira": {
            "token": "AAA...",
            "base_url": "https://my-dummy-jira-url.com",
            "server_api_v2": "rest/api/2",
            "agile_api_v1": "rest/agile/1.0",
            "greenhopper_api": "rest/greenhopper/1.0",
        },
        "retries": {"tries": 10, "delay": 0.0, "backoff": 2},
        "requests": {"timeout": 60},
        "concurrency": {"semaphore_limit": 50},
        "pagination": {"page_results": 40},
    }

    __sections: ClassVar = ["jira", "log", "retries", "requests", "concurrency"]

    @staticmethod
    def configuration(
        section: str, setting: str
    ) -> str | float | dict | logging.Logger | asyncio.BoundedSemaphore:
        """
        Retrieves the specified configuration.

        Method returns the configuration setting for the provided
        section and setting

        Args:
            section (str): Section of the configuration.
            setting (str): Specific setting of the configuration under the section.

        Returns:
            Value of the configuration requested.
        """
        return YajawConfig._configuration_settings[section][setting]

    @staticmethod
    def update_configuration(
        section: str,
        setting: str,
        value: str | float | dict | logging.Logger | asyncio.BoundedSemaphore,
    ):
        """
        Update the provided configuration setting.

        Updates the configuration dictionary with the informed setting value
        using the section and setting keys and refresh the class attributes.

        Args:
            section (str): Section of the configuration.
            setting (str): Specific setting of the configuration under the section.\
            value (str | float | dict | logging.Logger | asyncio.BoundedSemaphore):\
            The setting value to be stored.

        Raises:
            NameError: Can't update the configuration using the provided section and settings.
        """
        if section in YajawConfig.__sections:
            YajawConfig._configuration_settings[section][setting] = value
            YajawConfig._set_class_variables()
        else:
            raise NameError

    @staticmethod
    def load_initial_settings():
        """
        Load settings from the configuration file.

        Yajaw settings are loaded to a dictionary in memory from a configuration file, or
        use default values if the file is missing. Jira instance and access must be updated
        if no file is found.
        """
        fname = "yajaw.toml"
        try:
            with open(Path.home() / ".yajaw" / fname, "rb") as toml:
                YajawConfig._configuration_settings = tomllib.load(toml)
        except FileNotFoundError:
            ...
        YajawConfig._configuration_settings["log"] = {}
        YajawConfig._configuration_settings["log"]["logger"] = YajawConfig._define_logger()
        limit = YajawConfig._configuration_settings["concurrency"]["semaphore_limit"]
        # Ensures da minimum value of 5 for the BoundedSemaphore
        semaphore_limit = (
            limit if limit > YajawConfig._MIN_SEMAPHORE_LIMIT else YajawConfig._MIN_SEMAPHORE_LIMIT
        )
        YajawConfig._configuration_settings["concurrency"]["semaphore"] = asyncio.BoundedSemaphore(
            semaphore_limit
        )
        YajawConfig._configuration_settings["pagination"]["default"] = {
            "startAt": 0,
            "maxResults": YajawConfig._configuration_settings["pagination"]["page_results"],
        }
        YajawConfig._set_class_variables()

    @staticmethod
    def _define_logger() -> logging.Logger:
        """
        Configures the logging.Logger used by yajaw.

        As a library, logger named 'yajaw' is set up but no additional configuration
        is made to display the messages. NullHandler adapter is set to this logger.
        Configure logging from the application. Extra parameter 'context_id' is set.

        Returns:
            The yajaw logging.Logger object.
        """

        logger = logging.getLogger("yajaw")
        logger.setLevel(logging.CRITICAL)
        ch = logging.NullHandler()
        logger.addHandler(ch)
        return logger

    @staticmethod
    def _set_class_variables():
        """
        Updates the class variablers with the setting from the dictionary.

        Yajaw settings are loaded to a dictionary in memory from a configuration file, or
        default values if the file is missing. The method takes the content from the
        dictionary and updates the class variables accordingly.
        """
        YajawConfig.JIRA_PAT = YajawConfig._configuration_settings["jira"]["token"]
        YajawConfig.JIRA_BASE_URL = YajawConfig._configuration_settings["jira"]["base_url"]
        YajawConfig.SERVER_API_V2 = YajawConfig._configuration_settings["jira"]["server_api_v2"]
        YajawConfig.SERVER_API = YajawConfig.SERVER_API_V2
        YajawConfig.AGILE_API_V1 = YajawConfig._configuration_settings["jira"]["agile_api_v1"]
        YajawConfig.AGILE_API = YajawConfig.AGILE_API_V1
        YajawConfig.GREENHOPPER_API = YajawConfig._configuration_settings["jira"]["greenhopper_api"]
        YajawConfig.TRIES = YajawConfig._configuration_settings["retries"]["tries"]
        YajawConfig.DELAY = YajawConfig._configuration_settings["retries"]["delay"]
        YajawConfig.BACKOFF = YajawConfig._configuration_settings["retries"]["backoff"]
        YajawConfig.LOGGER = YajawConfig._configuration_settings["log"]["logger"]
        YajawConfig.SEMAPHORE = YajawConfig._configuration_settings["concurrency"]["semaphore"]
        YajawConfig.TIMEOUT = YajawConfig._configuration_settings["requests"]["timeout"]
        YajawConfig.DEFAULT_PAGINATION = YajawConfig._configuration_settings["pagination"][
            "default"
        ]



"""
Module responsible for the definition of yajaw exceptions.
"""


class YajawError(Exception):
    """
    Base class for yajaw errors.
    Error is derived from super class Exception.
    """


class InvalidResponseError(YajawError):
    """
    Response received is not valid. It is usually raised from
    underlying problems, such as runtime errors in HTTP requests,
    or processing the received results.
    Error is derived from super class YajawError.
    """


class HttpClientError(YajawError):
    """
    Base class for errors associated with a HTTP Status Code 4xx.
    Error is derived from super class YajawError.
    """


class HttpServerError(YajawError):
    """
    Base class for errors associated with a HTTP Status Code 5xx.
    Error is derived from super class HttpClientError.
    """


class ResourceUnauthorizedError(HttpClientError):
    """
    Authenticated user is unauthorized from access the resource.
    Error is derived from super class HttpClientError.
    """


class ResourceForbiddenError(HttpClientError):
    """
    Authenticated user is forbidden from access the resource.
    Error is derived from super class HttpClientError.
    """


class ResourceNotFoundError(HttpClientError):
    """
    Requested resource could not be found as informed.
    Error is derived from super class HttpClientError.
    """


class ResourceMethodNotAllowedError(HttpClientError):
    """
    Requested resource does not support the informed HTTP method.
    Low level error for internal modules use.
    Error is derived from super class HttpClientError.
    """



"""Module responsible for lower level HTTP requests."""
import asyncio
import math
import secrets
from http import HTTPStatus

import httpx

from yajaw import Option, YajawConfig, exceptions


class _PersonalAccessTokenAuth(httpx.Auth):
    """Class used for Personal Access Token auth in httpx."""

    def __init__(self, pat):
        self.token = pat

    def auth_flow(self, request):
        request.headers["Authorization"] = f"Bearer {self.token}"
        yield request


def _generate_headers() -> dict:
    """Function responsible for generating the headers info for HTTP requests."""
    return {"Accept": "application/json", "Content-Type": "application/json"}


def _generate_params(new_params: dict, existing_params: dict | None = None) -> dict:
    """Function responsible for generating the parameters info for HTTP requests."""
    if existing_params is None:
        existing_params = {}
    return existing_params | new_params


def _generate_payload(new_content: dict, existing_content: dict | None = None) -> dict:
    """Function responsible for generating the payload info for HTTP requests."""
    if existing_content is None:
        existing_content = {}
    return existing_content | new_content


def _generate_url(resource: str, api: str) -> str:
    """Function responsible for generating the url info for HTTP requests."""
    return f"{YajawConfig.JIRA_BASE_URL}/{api}/{resource}"


def _generate_auth() -> _PersonalAccessTokenAuth:
    """Function responsible for generating the authentication info for HTTP requests."""
    return _PersonalAccessTokenAuth(YajawConfig.JIRA_PAT)


def _generate_client() -> httpx.AsyncClient:
    """Function responsible for generating the client used in the context for HTTP requests."""
    return httpx.AsyncClient(
        auth=_generate_auth(),
        headers=_generate_headers(),
        timeout=YajawConfig.TIMEOUT,
        follow_redirects=True,
    )


class JiraInfo:
    """
    Class representing the Jira instance and the necessary parameters for HTTP requests.
    """

    def __init__(
        self,
        method: str,
        resource: str,
        api: str,
        params: dict | None = None,
        payload: dict | None = None,
    ):
        """
        Initializes a JiraInfo object with the given configuration.

        Args:
            method (str): HTTP method to be used for the request.
            resource (str): Resource part of the URL endpoint.
            api (str): API part of the URL endpoint.
            params (Optional[dict]): Parameters for the HTTP request; defaults to None.
            payload (Optional[dict]): Payload for the HTTP request; defaults to None.
        """
        self.method = method
        self.resource = resource
        self.api = api
        self.url = _generate_url(resource, api)
        self.params = params or {}
        self.payload = payload or {}


def _retry_response_error_detected(result: httpx.Response) -> bool:
    """Check if a retry should proceed based on the HTTP response."""
    retry = True
    if not isinstance(result, httpx.Response):
        _log_and_raise_request_issue("InvalidResponseError")
    if result.status_code == HTTPStatus.UNAUTHORIZED:
        _log_and_raise_request_issue("ResourceUnauthorizedError")
    if result.status_code == HTTPStatus.FORBIDDEN:
        _log_and_raise_request_issue("ResourceForbiddenError")
    if result.status_code == HTTPStatus.METHOD_NOT_ALLOWED:
        _log_and_raise_request_issue("ResourceMethodNotAllowedError")
    if result.status_code == HTTPStatus.NOT_FOUND:
        _log_and_raise_request_issue("ResourceNotFoundError")
    if httpx.codes.is_success(result.status_code):
        retry = False
    return retry


def _log_and_raise_request_issue(error_type: str, warning: Option = Option.NO) -> None:
    """Log an error or warning message and raise the specified exception."""
    log_level = YajawConfig.LOGGER.warning if warning is Option.YES else YajawConfig.LOGGER.error
    log_message = f"log message -- {error_type}"
    log_level(log_message)
    raise getattr(exceptions, error_type)


async def _retry_request(jira: JiraInfo, client: httpx.AsyncClient):
    """Retry the given function on certain conditions."""
    delay = secrets.SystemRandom().uniform(0, YajawConfig.DELAY)
    for attempt in range(1, YajawConfig.TRIES + 1):
        await asyncio.sleep(delay)
        result = await _send_request(jira=jira, client=client)
        _log_attempt_info(result, attempt, delay, error=Option.NO)
        if not _retry_response_error_detected(result):
            return result
        delay *= YajawConfig.BACKOFF
    _log_attempt_info(result, attempt, delay, error=Option.YES)
    raise exceptions.InvalidResponseError


def _log_attempt_info(result, attempt, delay, error=Option.NO):
    """Log information for a retry attempt."""
    log_level = YajawConfig.LOGGER.error if error == Option.YES else YajawConfig.LOGGER.info

    if isinstance(result, httpx.Response):
        log_message = (
            f"{result.status_code} - attempt {attempt:02d} - "
            f"delay {delay:05.2f}s -- {result.request.method} "
            f"{result.request.url}"
        )
    else:
        log_message = f"Non-HTTPX response - attempt {attempt:02d} - delay {delay:04.2f}s"

    log_level(log_message)


async def _send_request(jira: JiraInfo, client: httpx.AsyncClient) -> httpx.Response:
    """Function responsible for making a low-level HTTP request."""
    method, url, params, payload = jira.method, jira.url, jira.params, jira.payload
    async with YajawConfig.SEMAPHORE:
        return await client.request(method=method, url=url, params=params, json=payload)


async def send_single_request(
    jira: JiraInfo, client: httpx.AsyncClient | None = None
) -> httpx.Response:
    """
    Sends a HTTP request to a Jira instance.

    Args:
        jira (JiraInfo): Object of JiraInfo class representing the Jira instance.
        client (httpx.AsyncClient | None, optional): A client object used in the HTTP request.
        It may be already created when the function is called by send_paginated_requests. A new
        client is created otherwise.

    Raises:
        exceptions.ResourceNotFoundError: Resource could not be found as informed.
        exceptions.YajawError: An error happened on a HTTP request.

    Returns:
        httpx.Response: _description_
    """
    try:
        if client is None:
            client = _generate_client()
        task = asyncio.create_task(_retry_request(jira=jira, client=client))
        response = await task
    except exceptions.ResourceNotFoundError as exc:
        YajawConfig.LOGGER.warning("Resource could not be found.")
        raise exceptions.ResourceNotFoundError from exc
    except exceptions.YajawError as e:
        YajawConfig.LOGGER.exception("An error happened on a HTTP request.")
        raise exceptions.YajawError from e
    return response


async def send_paginated_requests(jira: JiraInfo) -> list[httpx.Response]:
    """
    Sends a paginated HTTP request to a Jira instance.

    Args:
        jira (JiraInfo): Object of JiraInfo class representing the Jira instance.

    Returns:
        list[httpx.Response]: List of response objects received from the requested pages.
    """
    default_page_attr = YajawConfig.DEFAULT_PAGINATION
    jira_list = _create_jira_list_with_page_attr(page_attr_list=[default_page_attr], jira=jira)
    initial_jira = jira_list[0]

    client = _generate_client()
    async with client:
        # First request with default pagination
        response = await send_single_request(jira=initial_jira, client=client)

        responses = []
        responses.append(response)

        # Identify if additional requests are needed
        page_attr = _retrieve_pagination_attributes(response=response)
        if _is_pagination_required(page_attr=page_attr):
            # Generate the updated page_attr_list
            page_attr_list = _create_list_of_page_attr(page_attr=page_attr)

            # Generate the updated jira_list
            jira_list = _create_jira_list_with_page_attr(page_attr_list=page_attr_list, jira=jira)

            # Create concurrent requests for the additional pages
            # iterate over jira_list
            # Consolidate responses list
            # for response in paginated_responses: responses.extend(response)

            tasks = [
                asyncio.create_task(send_single_request(jira=jira, client=client))
                for jira in jira_list
            ]
            group_of_response = await asyncio.gather(*tasks)

            responses.extend(list(group_of_response))

    return responses


def _create_list_of_page_attr(page_attr: dict) -> list[dict]:
    """Function that generates a list of attributes needed to retrieve all pages."""
    if "total" not in page_attr:
        return [page_attr]

    page_attr["start_at"] = page_attr["max_results"]

    page_attr_list = [
        page_attr["start_at"] + i * page_attr["max_results"]
        for i in range(int(math.ceil(page_attr["total"] / page_attr["max_results"])) - 1)
    ]
    return [{"startAt": page, "maxResults": page_attr["max_results"]} for page in page_attr_list]


# Assuming _generate_params and _generate_payload behave as before, merging dictionaries


def _create_jira_list_with_page_attr(page_attr_list: list[dict], jira: JiraInfo) -> list[JiraInfo]:
    """
    Function that gets a list of page attributes and creates
    a list with respective JiraInfo objects.
    """
    jira_list = []
    for page_attr in page_attr_list:
        if jira.method == "GET":
            params = _generate_params(new_params=page_attr, existing_params=jira.params)
            new_jira = JiraInfo(
                method=jira.method,
                resource=jira.resource,
                api=jira.api,
                params=params,
                payload=jira.payload,
            )
        elif jira.method == "POST":
            payload = _generate_payload(new_content=page_attr, existing_content=jira.payload)
            new_jira = JiraInfo(
                method=jira.method,
                resource=jira.resource,
                api=jira.api,
                params=jira.params,
                payload=payload,
            )
        else:
            # Handle other HTTP methods or raise an error
            new_jira = jira

        jira_list.append(new_jira)
    return jira_list


def _retrieve_pagination_attributes(response: httpx.Response) -> dict:
    """Function that gets the pagination valuesa from response and returns a dict with them"""
    page_attr = {}
    response_json = response.json()
    page_attr["start_at"] = response_json["startAt"] if "startAt" in response_json else None
    page_attr["max_results"] = (
        response_json["maxResults"] if "maxResults" in response_json else None
    )
    page_attr["total"] = response_json["total"] if "total" in response_json else None
    return page_attr


def _is_pagination_required(page_attr: dict) -> bool:
    """Function that checks if pagination is required."""
    start_at = page_attr["start_at"]
    max_results = page_attr["max_results"]
    total = page_attr["total"]
    return start_at + max_results < total



"""TBD"""
from http import HTTPStatus
from unittest.mock import MagicMock, patch

import httpx
import pytest

from yajaw.configuration import YajawConfig
from yajaw.core.rest import _retry_request, _retry_response_error_detected
from yajaw.exceptions import (
    InvalidResponseError,
    ResourceForbiddenError,
    ResourceMethodNotAllowedError,
    ResourceNotFoundError,
    ResourceUnauthorizedError,
)


@pytest.mark.asyncio
async def test_retry_response_error_detected():
    """TBD"""
    # Test case where result is not an instance of httpx.Response
    with pytest.raises(InvalidResponseError):
        _retry_response_error_detected("not_a_response")

    # Test case where result has status code UNAUTHORIZED
    result = MagicMock(spec=httpx.Response, status_code=HTTPStatus.UNAUTHORIZED)
    with pytest.raises(ResourceUnauthorizedError):
        _retry_response_error_detected(result)

    # Test case where result has status code FORBIDDEN
    result = MagicMock(spec=httpx.Response, status_code=HTTPStatus.FORBIDDEN)
    with pytest.raises(ResourceForbiddenError):
        _retry_response_error_detected(result)

    # Test case where result has status code METHOD_NOT_ALLOWED
    result = MagicMock(spec=httpx.Response, status_code=HTTPStatus.METHOD_NOT_ALLOWED)
    with pytest.raises(ResourceMethodNotAllowedError):
        _retry_response_error_detected(result)

    # Test case where result has status code NOT_FOUND
    result = MagicMock(spec=httpx.Response, status_code=HTTPStatus.NOT_FOUND)
    with pytest.raises(ResourceNotFoundError):
        _retry_response_error_detected(result)

    # Test case where result has a different status code
    result = MagicMock(spec=httpx.Response, status_code=HTTPStatus.OK)
    assert not _retry_response_error_detected(result)


@pytest.mark.asyncio
async def test_retry_request():
    """TBD"""
    jira_info = MagicMock()
    client = MagicMock(spec=httpx.AsyncClient)
    YajawConfig.TRIES = 3
    YajawConfig.DELAY = 1

    # Mock _send_request function
    with pytest.raises(InvalidResponseError), patch(
        "yajaw.core.rest._send_request", return_value=MagicMock("not a response object")
    ) as mock_send_request:
        await _retry_request(jira_info, client)

    mock_send_request.assert_called_with(jira=jira_info, client=client)



# SPDX-FileCopyrightText: 2023-present rmrighes <rmrighes@users.noreply.github.com>
#
# SPDX-License-Identifier: MIT
"""
File __about__.py responsible for yajaw package meta information.
"""


__version__ = "0.1.0.dev1"



# SPDX-FileCopyrightText: 2023-present rmrighes <rmrighes@users.noreply.github.com>
#
# SPDX-License-Identifier: MIT
"""File __init__.py responsible for enabling the import of yajaw.utils package."""


__all__ = ["performance", "concurrency"]



"""
Module wrapping up the supported Jira resources.
It is the main external interface for yajaw users.
"""
import asyncio

from yajaw import ApiType, YajawConfig
from yajaw import exceptions as e
from yajaw.core import rest
from yajaw.utils.concurrency import async_to_sync


async def async_fetch_all_projects(expand: str | None = None) -> list[dict]:
    """
    Async call to fetch all projects.

    It is intended to be used on asynchronous code. Use the sync version otherwise.
    Returns all projects visible to the user. It is based on the API
    GET /rest/api/2/project. That resource is deprecated and will be replaced
    by GET /rest/api/2/project/search, which will support pagination.


    Args:
        expand (str | None, optional): Expect a simple string with a comma-separated\
        list of attributes to be expanded.\
        They are: description, issueTypes, lead, and projectKeys.

    Returns:
        List of dictionaries representing the returned projects.\
        An empty list is returned if nothing found.
    """
    expand_dict = {} if expand is None else {"expand": expand}

    jira = rest.JiraInfo(
        method="GET",
        resource="project",
        api=YajawConfig.SERVER_API,
        params=expand_dict,
        payload=None,
    )

    try:
        response = await rest.send_single_request(jira=jira)
        return response.json()
    except e.ResourceNotFoundError:
        return []


@async_to_sync
def fetch_all_projects(expand: str | None = None) -> list[dict]:
    """
    Sync call to fetch all projects.

    It is intended to be used on synchronous code. Use the async version otherwise.
    Returns all projects visible to the user. It is based on the API
    GET /rest/api/2/project. That resource is deprecated and will be replaced
    by GET /rest/api/2/project/search, which will support pagination.


    Args:
        expand (str | None, optional): Expect a simple string with a comma-separated\
        list of attributes to be expanded.\
        They are: description, issueTypes, lead, and projectKeys.

    Returns:
        List of dictionaries representing the returned projects.\
        An empty list is returned if nothing found.
    """
    return async_fetch_all_projects(expand=expand)


async def async_fetch_project(project_key: str, expand: str | None = None) -> dict:
    """
    Async call to fetch the details of a single project.

     It is intended to be used on asynchronous code. Use the sync version otherwise.
     Return the details of a single project. It is based on the API
     GET /rest/api/2/project/{projectKey}.

    Args:
        project_key (str): Key identifier for the project.
        expand (str | None, optional): Expect a simple string with a comma-separated\
        list of attributes to be expanded. They are: description, issueTypes, lead,\
        projectKeys, and issueTypeHierarchy. Defaults to None.

    Returns:
        Dictionary with the project details. An empty dictionary is returned\
        if nothing is found.
    """
    expand_dict = {} if expand is None else {"expand": expand}

    jira = rest.JiraInfo(
        method="GET",
        resource=f"project/{project_key}",
        api=YajawConfig.SERVER_API,
        params=expand_dict,
        payload=None,
    )

    try:
        response = await rest.send_single_request(jira=jira)
        return response.json()
    except e.ResourceNotFoundError:
        return {}


@async_to_sync
def fetch_project(project_key: str, expand: str | None = None) -> dict:
    """
    Sync call to fetch the details of a single project.

     It is intended to be used on synchronous code. Use the async version otherwise.
     Return the details of a single project. It is based on the API
     GET /rest/api/2/project/{projectKey}.

    Args:
        project_key (str): Key identifier for the project.
        expand (str | None, optional): Expect a simple string with a comma-separated\
        list of attributes to be expanded. They are: description, issueTypes, lead,\
        projectKeys, and issueTypeHierarchy. Defaults to None.

    Returns:
        Dictionary with the project details. An empty dictionary is returned\
        if nothing is found.
    """
    return async_fetch_project(project_key=project_key, expand=expand)


async def async_fetch_projects_from_list(
    project_keys: list[str], expand: str | None = None
) -> list[dict]:
    """
    Async call to fetch the details of a list of projects.

    It is intended to be used on asynchronous code. Use the sync version otherwise.
    Return the details of a list of projects. It is based on the API
    GET /rest/api/2/project/{projectKey} and the function calls it as many times as
    items in the provided list.

    Args:
        project_keys (list[str]): A list of strings representing the project keys.
        expand (str | None, optional): Expect a simple string with a comma-separated\
        list of attributes to be expanded. They are: description, issueTypes, lead,\
        projectKeys, and issueTypeHierarchy. Defaults to None.

    Returns:
        List of dictionaries representing the returned projects.\
        An empty list is returned if nothing found.
    """
    expand_dict = {} if expand is None else {"expand": expand}

    jira_list = [
        rest.JiraInfo(
            method="GET",
            resource=f"project/{project_key}",
            api=YajawConfig.SERVER_API,
            params=expand_dict,
            payload=None,
        )
        for project_key in project_keys
    ]

    try:
        tasks = [asyncio.create_task(rest.send_single_request(jira=jira)) for jira in jira_list]
        responses = await asyncio.gather(*tasks)
        return [response.json() for response in responses]
    except e.ResourceNotFoundError:
        return []


@async_to_sync
def fetch_projects_from_list(project_keys: list[str], expand: str | None = None) -> list[dict]:
    """
    Sync call to fetch the details of a list of projects.

    It is intended to be used on synchronous code. Use the async version otherwise.
    Return the details of a list of projects. It is based on the API
    GET /rest/api/2/project/{projectKey} and the function calls it as many times as
    items in the provided list.

    Args:
        project_keys (list[str]): A list of strings representing the project keys.
        expand (str | None, optional): Expect a simple string with a comma-separated\
        list of attributes to be expanded. They are: description, issueTypes, lead,\
        projectKeys, and issueTypeHierarchy. Defaults to None.

    Returns:
        List of dictionaries representing the returned projects.\
        An empty list is returned if nothing found.
    """
    return async_fetch_projects_from_list(project_keys=project_keys, expand=expand)


async def async_fetch_issue(
    issue_key: str, expand: str | None = None, api: ApiType = ApiType.CLASSIC
) -> dict:
    """
    Async call to fetch the details of a single issue.

    It is intended to be used on asynchronous code. Use the sync version otherwise.
    Return the details of a single issue. It is based on the API
    GET /rest/api/2/issue/{issueKey}. If the identifier doesn't match an issue,
    a case-insensitive search and check for moved issues is performed.
    If a matching issue is found its details are returned.

    Args:
        issue_key (str):  Key identifier for the issue.
        expand (str | None, optional): Expect a simple string with a comma-separated\
        list of attributes to be expanded. They are: renderedFields, name, schema,\
        transitions,editmeta, changelog, and versionedRepresentations.\
        Defaults to None.
        api (ApiType, optional): Takes the enumeration values ApiType.CLASSIC or\
        ApiType.AGILE as possible values to represent which API to be used. Defaults\
        to ApiType.CLASSIC.

    Returns:
        Dictionary with the issue details. An empty dictionary is returned\
        if nothing is found.
    """
    expand_dict = {} if expand is None else {"expand": expand}

    api_value = YajawConfig.SERVER_API if api == ApiType.CLASSIC else YajawConfig.AGILE_API

    jira = rest.JiraInfo(
        method="GET", resource=f"issue/{issue_key}", api=api_value, params=expand_dict, payload=None
    )

    try:
        response = await rest.send_single_request(jira=jira)
        return response.json()
    except e.ResourceNotFoundError:
        return {}


@async_to_sync
def fetch_issue(issue_key: str, expand: str | None = None, api: ApiType = ApiType.CLASSIC) -> dict:
    """
    Sync call to fetch the details of a single issue.

    It is intended to be used on synchronous code. Use the async version otherwise.
    Return the details of a single issue. It is based on the API
    GET /rest/api/2/issue/{issueKey}. If the identifier doesn't match an issue,
    a case-insensitive search and check for moved issues is performed.
    If a matching issue is found its details are returned.

    Args:
        issue_key (str):  Key identifier for the issue.
        expand (str | None, optional): Expect a simple string with a comma-separated\
        list of attributes to be expanded. They are: renderedFields, name, schema,\
        transitions,editmeta, changelog, and versionedRepresentations.\
        Defaults to None.
        api (ApiType, optional): Takes the enumeration values ApiType.CLASSIC or\
        ApiType.AGILE as possible values to represent which API to be used. Defaults\
        to ApiType.CLASSIC.

    Returns:
        Dictionary with the issue details. An empty dictionary is returned\
        if nothing is found.
    """
    return async_fetch_issue(issue_key=issue_key, expand=expand, api=api)


async def async_search_issues(jql: str, expand: str | None = None) -> list[dict]:
    """
    Async call to fetch the result of a search for issues using JQL.

    It is intended to be used on asynchronous code. Use the sync version otherwise.
    Return the results of a search for issues using the provided JQL. It is based on
    the API POST /rest/api/2/search.

    Args:
        jql (str): A valid Jira Query Language in string format.
        expand (str | None, optional): Expect a simple string with a comma-separated\
        list of attributes to be expanded. They are: renderedFields, name, schema,\
        transitions, operations, editmeta, changelog, and versionedRepresentations.\
        Defaults to None.

    Returns:
        List of dictionaries representing the returned projects.\
        An empty list is returned if nothing found.
    """
    expand_dict = {} if expand is None else {"expand": expand}
    query = {"jql": jql}

    jira = rest.JiraInfo(
        method="POST",
        resource="search",
        api=YajawConfig.SERVER_API,
        params=expand_dict,
        payload=query,
    )

    try:
        responses = await rest.send_paginated_requests(jira=jira)
        return [issue for issue_page in responses for issue in issue_page.json()["issues"]]
    except e.ResourceNotFoundError:
        return []


@async_to_sync
def search_issues(jql: str, expand: str | None = None) -> list[dict]:
    """
    Sync call to fetch the result of a search for issues using JQL.

    It is intended to be used on synchronous code. Use the async version otherwise.
    Return the results of a search for issues using the provided JQL. It is based on
    the API POST /rest/api/2/search.

    Args:
        jql (str): A valid Jira Query Language in string format.
        expand (str | None, optional): Expect a simple string with a comma-separated\
        list of attributes to be expanded. They are: renderedFields, name, schema,\
        transitions, operations, editmeta, changelog, and versionedRepresentations.\
        Defaults to None.

    Returns:
        List of dictionaries representing the returned projects.\
        An empty list is returned if nothing found.
    """
    return async_search_issues(jql=jql, expand=expand)



"""
Module responsible for utilitarian decorators used to run async function
synchronously.
"""
import asyncio
import concurrent.futures
import functools
import threading


def async_to_sync(func):
    """
    async_to_sync Decorator used to run async function synchronously.

    Decorated target functions with this one will determine the right
    approach to execute the computation of asynchronous code, such as
    running in thread safe mode, or managing the event loop to be used.
    """

    async def main_wrap(args, kwargs, call_result):
        "Main wrap logic that consists on running the async wrapped function."
        try:
            result = await func(*args, **kwargs)
        except asyncio.CancelledError as e:
            call_result.set_exception(e)
        else:
            call_result.set_result(result)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        "Function responsible for how to execute the asynchronous computation."

        def run_in_thread():
            "Set up a new event loop and run it the until the main_wrap function is done."
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try:
                loop.run_until_complete(main_wrap(args, kwargs, call_result))
            finally:
                loop.close()

        call_result = concurrent.futures.Future()
        threadlocal = False

        try:
            main_event_loop = asyncio.get_running_loop()

        except RuntimeError:
            main_event_loop = getattr(threading.local(), "main_event_loop", None)
            threadlocal = True

        # Check for NoneType first so the second clause does not fail
        if main_event_loop and main_event_loop.is_running():
            if threadlocal:
                # Data is thread specific
                main_event_loop.call_soon_threadsafe(
                    main_event_loop.create_task, main_wrap(args, kwargs, call_result)
                )

            else:
                thread = threading.Thread(target=run_in_thread)
                thread.start()
                thread.join()

        else:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try:
                loop.run_until_complete(main_wrap(args, kwargs, call_result))
            finally:
                try:
                    if hasattr(loop, "shutdown_asyncgens"):
                        loop.run_until_complete(loop.shutdown_asyncgens())
                finally:
                    loop.close()
                    asyncio.set_event_loop(main_event_loop)

        return call_result.result()

    return wrapper
